# Cryomagnetics Model 4G Superconducting Magnet Power Supply
# Created by Dacen Waters for Yankowitz Lab VTI
# Last updated June 1st, 2021

import logging
from time import sleep

from qcodes import VisaInstrument
from qcodes import validators as vals

log = logging.getLogger(__name__)

DEFAULT_LIMITS = {
    "RANGE0": 40.0,  # A
    "RANGE1": 80.0,
    "RANGE2": 96.5,
    "RANGE3": 98.0,
    "RANGE4": 99.0,
}
DEFAULT_RATES = {
    "RANGE0": 0.04,  # A/s
    "RANGE1": 0.02,
    "RANGE2": 0.01,
    "RANGE3": 0.001,
    "RANGE4": 0.001,
}
MAX_RATES = {
    "RANGE0": 0.0468,  # A/s
    "RANGE1": 0.0234,
    "RANGE2": 0.0117,
    "RANGE3": 0.001,
    "RANGE4": 0.001,
}

MAX_SUPPLY_CURRENT = 100  # A
FIELD_TO_CURRENT = 1246.7 / 1e4  # T/A

MIN_FIELD = -12.0  # T
MAX_FIELD = 12.0


class M4G(VisaInstrument):
    def __init__(self, name, address, **kwargs):
        super().__init__(name, address, terminator="\r\n", **kwargs)

        self.add_parameter(
            name="iout", unit="A", get_cmd=self._get_iout, docstring="Output current"
        )

        self.add_parameter(
            name="vout", unit="V", get_cmd=self._get_vout, docstring="Output voltage"
        )
        self.add_parameter(
            name="units",
            get_cmd="UNITS?",
            set_cmd="UNITS {}",
            docstring="Units",
            vals=vals.Enum("A", "kG", "T"),
        )

        self.add_parameter(
            name="ulim",
            get_cmd=self._get_ulim,
            set_cmd="ULIM {}",
            docstring="Upper Limit",
            get_parser=float,
        )
        self.add_parameter(
            name="llim",
            get_cmd=self._get_llim,
            set_cmd="LLIM {}",
            docstring="Lower Limit",
            get_parser=float,
        )
        self.add_parameter(
            name="field",
            get_cmd=self._get_field,
            set_cmd=self._set_field,
            unit="T",
            docstring="Magnetic Field",
            vals=vals.Numbers(MIN_FIELD, MAX_FIELD),
        )

        self.add_parameter(
            "range0_rate",
            unit="A/s",
            get_cmd="RATE? 0",
            set_cmd=self._set_range0_rate,
            get_parser=float,
            docstring="Ramp rate for Range 0",
            vals=vals.Numbers(0, MAX_SUPPLY_CURRENT),
        )
        self.add_parameter(
            "range1_rate",
            unit="A/s",
            get_cmd="RATE? 1",
            set_cmd=self._set_range1_rate,
            get_parser=float,
            docstring="Ramp rate for Range 1",
            vals=vals.Numbers(0, MAX_SUPPLY_CURRENT),
        )
        self.add_parameter(
            "range2_rate",
            unit="A/s",
            get_cmd="RATE? 2",
            set_cmd=self._set_range2_rate,
            get_parser=float,
            docstring="Ramp rate for Range 2",
            vals=vals.Numbers(0, MAX_SUPPLY_CURRENT),
        )
        self.add_parameter(
            "range3_rate",
            unit="A/s",
            get_cmd="RATE? 3",
            set_cmd=self._set_range3_rate,
            get_parser=float,
            docstring="Ramp rate for Range 3",
            vals=vals.Numbers(0, MAX_SUPPLY_CURRENT),
        )
        self.add_parameter(
            "range4_rate",
            unit="A/s",
            get_cmd="RATE? 4",
            set_cmd=self._set_range4_rate,
            get_parser=float,
            docstring="Ramp rate for Range 4",
            vals=vals.Numbers(0, MAX_SUPPLY_CURRENT),
        )
        self.add_parameter(
            "range0_limit",
            unit="A",
            get_cmd="RANGE? 0",
            set_cmd=self._set_range0_limit,
            get_parser=float,
            docstring="Ramp limit for Range 0",
            vals=vals.Numbers(0, MAX_SUPPLY_CURRENT),
        )
        self.add_parameter(
            "range1_limit",
            unit="A",
            get_cmd="RANGE? 1",
            set_cmd=self._set_range1_limit,
            get_parser=float,
            docstring="Ramp limit for Range 1",
            vals=vals.Numbers(0, MAX_SUPPLY_CURRENT),
        )
        self.add_parameter(
            "range2_limit",
            unit="A",
            get_cmd="RANGE? 2",
            set_cmd=self._set_range2_limit,
            get_parser=float,
            docstring="Ramp limit for Range 2",
            vals=vals.Numbers(0, MAX_SUPPLY_CURRENT),
        )
        self.add_parameter(
            "range3_limit",
            unit="A",
            get_cmd="RANGE? 3",
            set_cmd=self._set_range3_limit,
            get_parser=float,
            docstring="Ramp limit for Range 3",
            vals=vals.Numbers(0, MAX_SUPPLY_CURRENT),
        )
        self.add_parameter(
            "range4_limit",
            unit="A",
            get_cmd="RANGE? 4",
            set_cmd=self._set_range4_limit,
            get_parser=float,
            docstring="Ramp limit for Range 4",
            vals=vals.Numbers(0, MAX_SUPPLY_CURRENT),
        )

    def _get_iout(self):  # Output current in current units
        result = self.ask("IMAG?")
        units = self.units()
        if units == "A":
            return float(result[:-1])
        else:
            return float(result[:-2])

    def _get_vout(self):  # Output voltage in volts
        result = self.ask("VOUT?")
        return float(result[:-1])

    def _get_ulim(self):  # Get upper limit in current units
        result = self.ask("ULIM?")
        units = self.units()
        if units == "A":
            return float(result[:-1])
        else:
            return float(result[:-2])

    def _get_llim(self):  # Get lower limit in current units
        result = self.ask("LLIM?")
        units = self.units()
        if units == "A":
            return float(result[:-1])
        else:
            return float(result[:-2])

    def _get_field(self):  # Returns the field in Tesla (regardless of current units)
        result = self.ask("IMAG?")
        units = self.units()

        for i in range(
            10
        ):  # Sometimes the magnet returns the IMAG? query for some reason. If this happens, wait one second and try asking again
            try:
                float(result[:-2])
                break
            except:
                print("Communication error: returned value " + result)
                sleep(1.0)
                result = self.ask("IMAG?")

        if units == "A":
            return float(result[:-1]) * FIELD_TO_CURRENT
        else:
            return (float(result[:-2])) / 10

    def _set_field(self, field):  # Sweeps to the input field value (field in T)
        self.setpoint(field)
        if field == 0:
            self.write("SWEEP ZERO")
        else:
            self.write("SWEEP UP")
        self.log.info(f"Sweeping field to {field} T")

    #         sleep(0.05)
    #         while self.field() != field:
    #             sleep(0.05)
    #         sleep(0.05)

    # Commands for setting the sweep rates
    # Each command checks that the corresponding limit does the maximum rate (error is raised if it does)
    def _set_range0_rate(self, rate):
        check_rate(self.range0_limit(), rate)
        self.write_raw(f"RATE 0 {rate}")

    def _set_range1_rate(self, rate):
        check_rate(self.range1_limit(), rate)
        self.write_raw(f"RATE 1 {rate}")

    def _set_range2_rate(self, rate):
        check_rate(self.range2_limit(), rate)
        self.write_raw(f"RATE 2 {rate}")

    def _set_range3_rate(self, rate):
        check_rate(self.range3_limit(), rate)
        self.write_raw(f"RATE 3 {rate}")

    def _set_range4_rate(self, rate):
        check_rate(self.range4_limit(), rate)
        self.write_raw(f"RATE 4 {rate}")

    # Commands for setting the current limits for each range
    # Each command raises an error if values don't corresponding to Rate i-1 > Rate i > Rate i+1 for each i
    def _set_range0_limit(self, limit):
        upper_limit = float(self.ask("RANGE? 1"))
        if limit < upper_limit:
            self.write_raw(f"RANGE 0 {limit}")
        else:
            raise ValueError(
                f"Range 0 limit must be lower than Range1 limit ({upper_limit} A)"
            )

    def _set_range1_limit(self, limit):
        lower_limit = float(self.ask("RANGE? 0"))
        upper_limit = float(self.ask("RANGE? 2"))
        if limit > lower_limit and limit < upper_limit:
            self.write_raw(f"RANGE 1 {limit}")
        else:
            raise ValueError(
                f"Range 1 limit must be between Range 0 limit ({lower_limit} A)"
                + f" and Range 2 limit ({upper_limit} A)"
            )

    def _set_range2_limit(self, limit):
        lower_limit = float(self.ask("RANGE? 1"))
        upper_limit = float(self.ask("RANGE? 3"))
        if limit > lower_limit and limit < upper_limit:
            self.write_raw(f"RANGE 2 {limit}")
        else:
            raise ValueError(
                f"Range 2 limit must be between Range 1 limit ({lower_limit} A)"
                + f" and Range 3 limit ({upper_limit} A)"
            )

    def _set_range3_limit(self, limit):
        lower_limit = float(self.ask("RANGE? 2"))
        upper_limit = float(self.ask("RANGE? 4"))
        if limit > lower_limit and limit < upper_limit:
            self.write_raw(f"RANGE 3 {limit}")
        else:
            raise ValueError(
                f"Range 3 limit must be between Range 2 limit ({lower_limit} A)"
                + f" and Range 4 limit ({upper_limit} A)"
            )

    def _set_range4_limit(self, limit):
        lower_limit = float(self.ask("RANGE? 3"))
        if limit > lower_limit:
            self.write_raw(f"RANGE 4 {limit}")
        else:
            raise ValueError(
                f"Range 4 limit must be higher than Range 3 limit ({lower_limit} A)"
            )

    # Standard commands for switching between remote and local control
    def remote(self):
        self.write_raw("REMOTE")
        self.is_remote()  # Runs the the test above for remote connection, if not, throws an error to notify the user

    def local(self):
        self.write_raw("LOCAL")
        # Command to quickly test if the remote connection works

    def is_remote(self):
        init_val = float(self.ask("RATE? 5"))
        if init_val > 0.1:
            Delta = -0.1
        else:
            Delta = +0.1
        self.write_raw(f"RATE 5 {init_val + Delta}")
        test_val = self.ask("RATE? 5")
        try:
            test_val = float(test_val)
        except:
            print("Connection test error. RATE? 5 returned: ", test_val)
            print("Reattempting remote test.")
            sleep(1.0)
            self.is_remote()
        self.write_raw(f"RATE 5 {init_val}")
        if test_val != init_val:
            pass
        else:
            raise RemoteError(
                "Remote control not enabled (toggle the Local button on the front panel until the purple 'Local' text vanishes from the screen)"
            )

    def examine(self):
        for r in range(5):
            limit = float(self.ask(f"RANGE? {r}"))
            rate = float(self.ask(f"RATE? {r}"))
            print(f"Range {r}:" + f"Limit = {limit} A, " + f"Rate = {rate} A/s")
        print(f"Field: {self.field()} T")
        print("Sweep status: " + self.sweep_status())

    # This command sets all the rates and corresponding limits to the default values, which are hard coded above
    def set_defaults(self):
        for r in range(5):
            rate_ = DEFAULT_RATES[f"RANGE{r}"]
            limit_ = DEFAULT_LIMITS[f"RANGE{r}"]
            self.write_raw(f"RATE {r} " + f"{rate_}")
            self.write_raw(f"RANGE {r} " + f"{limit_}")
            sleep(0.1)

    # Checks sweep status
    def sweep_status(self):
        return self.ask("SWEEP?")

    # Function to establish the upper limit as the setpoint with the lower limit to be gauranteed to less than the upper limit
    def setpoint(self, field):  # field in Tesla
        units = self.units()
        llim_field = field - 0.1
        if units == "A":
            ulim_setpoint = field / FIELD_TO_CURRENT
            llim_setpoint = llim_field / FIELD_TO_CURRENT
        else:
            ulim_setpoint = field * 10
            llim_setpoint = llim_field * 10
        self.write(f"LLIM {llim_setpoint}")
        sleep(0.05)
        self.write(f"ULIM {ulim_setpoint}")

    # Sets the first three rates to a constant value, with the usual buffers for RANGES 3 and 4
    def constant_rate(self, rate):
        check_rate(
            DEFAULT_LIMITS["RANGE2"], rate
        )  # Desired rate can't be faster than the DEFAULT_LIMIT for RANGE 2
        for r in range(3):
            limit = DEFAULT_LIMITS[f"RANGE{r}"]
            self.write_raw(f"RATE {r} " + f"{rate}")
            self.write_raw(f"RANGE {r} " + f"{limit}")
            sleep(0.1)
        for r in range(3, 5):
            limit = DEFAULT_LIMITS[f"RANGE{r}"]
            rate_ = DEFAULT_RATES[f"RANGE{r}"]
            self.write_raw(f"RATE {r} " + f"{rate_}")
            self.write_raw(f"RANGE {r} " + f"{limit}")
            sleep(0.1)


# The following two commands are for testing that the input rates are not higher than the max rates, hard coded above
def find_range(current):
    current = abs(current)
    if current <= DEFAULT_LIMITS["RANGE0"]:
        return "RANGE0"
    if current >= DEFAULT_LIMITS["RANGE4"]:
        return "RANGE4"
    for i in range(1, 5):
        if DEFAULT_LIMITS[f"RANGE{i - 1}"] < current <= DEFAULT_LIMITS[f"RANGE{i}"]:
            return f"RANGE{i}"


def check_rate(current, rate):
    range_ = find_range(current)
    if rate > MAX_RATES[range_]:
        raise ValueError(
            "Rate is too high for desired current range. Check MAX_RATES and corresponding DEFAULT_LIMITS (Page 20 of Manual, Table 5-1)"
        )
    else:
        pass


class RemoteError(Exception):
    pass



# sweep1d.py

import time

from PyQt5.QtCore import QObject, pyqtSlot

from ..tools.util import safe_get
from .base_sweep import BaseSweep
from .progress import SweepState

try:
    from qcodes.instrument_drivers.american_magnetics import AMIModel430 as AMI430
except ImportError:  # pragma: no cover - fallback for older QCoDeS versions
    from qcodes.instrument_drivers.american_magnetics.AMI430 import AMI430
from functools import partial

from qcodes_contrib_drivers.drivers.OxfordInstruments.IPS120 import (
    OxfordInstruments_IPS120,
)

from ..Drivers.M4G import M4G


class Sweep1D(BaseSweep, QObject):
    """Class extending BaseSweep to sweep one independent parameter.

    The independent parameter defaults to time until a new parameter is
    assigned to 'set_param'. Sweep1D will sweep this parameter to its stop
    value (and back) while measuring any changes in the followed parameters.
    The data will be live-plotted and saved to a database if desired.

    Attributes:
    ---------
    set_param:
        QCoDeS Parameter to be swept.
    begin:
        The starting value of the sweeping parameter.
    end:
        The ending value of the sweeping parameter.
    step:
        The step size between measured datapoints.
    bidirectional:
        Determines whether or not to run the sweep in both directions.
    runner:
        Assigns the Runner Thread.
    plotter:
        Assigns the Plotter Thread.
    datasaver:
        Initiated by Runner Thread to enable saving and export of data.
    inter_delay:
        Time (in seconds) to wait between data points.
    save_data:
        Flag used to determine if the data should be saved or not.
    plot_data:
        Flag to determine whether or not to live-plot data.
    complete_func:
        Sets a function to be executed upon completion of the sweep.
    x_axis_time:
        Defaults to 0 to allow the 'set_param' to be plotted on the x-axis.
    parent:
        Sets a parent Sweep2D object.
    continual:
        Causes sweep to continuously run back and forth between start and stop points.
    plot_bin:
        Sets the number of data points taken between updates of the plot. Defaults to 1.
    back_multiplier:
        Factor to scale the step size after flipping directions.
    setpoint:
        The first parameter value where a measurement is obtained after starting.
    direction:
        Either 0 or 1 to indicate the direction of the sweep.
    ramp_sweep:
        Defines the sweep used to set the parameter to its starting value.
    instrument:
        The instrument used to control/monitor the 'set_param'.

    Methods:
    ---------
    start(persist_data=None, ramp_to_start=True, ramp_multiplier=1)
        Starts the sweep; runs from the BaseSweep start() function.
    pause()
        Pauses any currently active sweeps.
    kill()
        Ends the threads spawned by the sweep and closes any active plots.
    step_param()
        Iterates the parameter.
    step_AMI430()
        Used to control sweeps of AMI430 Instrument.
    step_IPS120()
        Used to control sweeps of IPS120 Instrument.
    flip_direction()
        Flips the direction of the sweep, to do bidirectional sweeping.
    ramp_to(value, start_on_finish=False, persist=None, multiplier=1)
        Ramps the set_param to a given value, at a rate dictated by the multiplier.
    ramp_to_zero()
        Ramps the set_param to 0, at the same rate as already specified.
    done_ramping(self, value, start_on_finish=False, pd=None)
        Alerts the sweep that the ramp is finished.
    get_param_setpoint()
        Obtains the current value of the setpoint.
    reset(new_params=None)
        Resets Sweep1D, optional argument to change sweep details.
    """

    def __init__(
        self,
        set_param,
        start,
        stop,
        step,
        bidirectional=False,
        continual=False,
        x_axis_time=0,
        err=1e-2,
        *args,
        **kwargs,
    ):
        """Initializes the sweep.

        There are 6 additional parameters not included in the BaseSweep initialization.

        Parameters
        ---------
        set param:
            The independent parameter to be swept.
        start:
            The value that the sweep will begin from.
        stop:
            The value that the sweep will end at.
        step:
            The step spacing for each measurement.
        x_axis_time:
            Determines if followed parameters are plotted against time(1) or 'set_param'(0).
        err:
            Tolerance for considering rounding errors when determining when the sweep has finished.
        """
        # Initialize the BaseSweep
        QObject.__init__(self)
        BaseSweep.__init__(
            self, set_param=set_param, x_axis_time=x_axis_time, *args, **kwargs
        )

        # Validate sweep range against parameter's validator bounds
        self._validate_param_sweep_range(set_param, start, stop)

        self.begin = start
        self.end = stop
        self.step = step

        # Store the start for snapping. Updated on direction flips to ensure
        # the step grid always includes the current sweep's starting point.
        self._snap_origin = start

        # Make sure the step is in the right direction
        if (self.end - self.begin) > 0:
            self.step = abs(self.step)
        else:
            self.step = (-1) * abs(self.step)

        self.setpoint = self.begin - self.step
        self.setpoint = self._snap_to_step(self.setpoint, self._snap_origin, self.step)
        self.bidirectional = bidirectional
        self.continuous = continual
        self.direction = 0
        self.ramp_sweep = None
        self.instrument = self.set_param.instrument
        self.err = err

        self.emit_step_info(
            self.set_param.label,
            self.begin,
            self.end,
            self.step,
            getattr(self.set_param, "unit", None),
        )

        self.magnet_initialized = False
        self._ami_completion_pending = False
        self._m4g_completion_pending = False
        if not self.continuous:
            self.progressState.progress = 0.0
            self.update_progress()

    def __str__(self):
        return f"1D Sweep of {self.set_param.label} from {self.begin} to {self.end}, with step size {self.step}."

    def __repr__(self):
        return f"Sweep1D({self.set_param.label}, {self.begin}, {self.end}, {self.step})"

    def start(self, persist_data=None, ramp_to_start=True, ramp_multiplier=1):
        """Starts the sweep; runs from the BaseSweep start() function.

        Parameters
        ---------
        persist_data:
            Used to set Sweep2D parameter.
        ramp_to_start:
            Gently sweeps parameter to desired starting value.
        ramp_multiplier:
            Factor to control ramping speed compared to sweep speed.
        """
        if self.progressState.state == SweepState.RAMPING:
            self.print_main.emit(
                "Still ramping. Wait until ramp is done to start the sweep."
            )
            return
        if self.progressState.state == SweepState.RUNNING:
            self.print_main.emit("Sweep is already running.")
            return

        self._ami_completion_pending = False
        self._m4g_completion_pending = False

        if (
            ramp_to_start is True
            and not isinstance(self.instrument, OxfordInstruments_IPS120)
            and not isinstance(self.instrument, AMI430)
            and not isinstance(self.instrument, M4G)
        ):
            self.print_main.emit(
                f"Ramping to our starting setpoint value of {self.begin} ({self.set_param.unit})"
            )
            self.ramp_to(
                self.begin,
                start_on_finish=True,
                persist=persist_data,
                multiplier=ramp_multiplier,
            )
        else:
            self.print_main.emit(
                f"Sweeping {self.set_param.label} to {self.end} ({self.set_param.unit})"
            )
            BaseSweep.start(self, persist_data=persist_data)

    def pause(self):
        """Pauses any currently active sweeps."""
        if self.progressState.state == SweepState.RAMPING and self.ramp_sweep is not None:
            self.print_main.emit("Stopping the ramp.")
            self.ramp_sweep.pause()
            self.ramp_sweep.kill()

            while self.ramp_sweep.check_running():
                time.sleep(0.2)
            self.done_ramping(self.ramp_sweep.setpoint)
            # self.setpoint=self.ramp_sweep.setpoint
            # self.ramp_sweep.plotter.clear()
            # self.ramp_sweep = None
            # self.progressState.state = SweepState.READY
            # self.print_main.emit(f"Stopped the ramp, the current setpoint  is {self.setpoint} {self.set_param.unit}")

        BaseSweep.pause(self)

        if isinstance(self.instrument, AMI430):
            self.instrument.pause()
            self.magnet_initialized = False
        elif isinstance(self.instrument, OxfordInstruments_IPS120):
            self.try_set(self.instrument.activity, 0)
            self.magnet_initialized = False
        elif isinstance(self.instrument, M4G):
            self.instrument.write("SWEEP PAUSE")

    def stop(self):
        """Stop/pause the sweep. Alias for pause() for backward compatibility.

        Handles stopping of ramp sweeps if in ramping state and instrument-specific pause operations.
        """
        self.pause()

    def kill(self):
        """Ends the threads spawned by the sweep and closes any active plots."""
        # Always tear down active ramp sweep, regardless of state
        # (error may have occurred during ramping, leaving ramp_sweep active)
        # Use getattr to handle case where __init__ failed before ramp_sweep was set
        if getattr(self, "ramp_sweep", None) is not None:
            try:
                self.ramp_sweep.pause()
            except Exception:
                pass
            try:
                self.ramp_sweep.kill()
            except Exception:
                pass
            self.ramp_sweep = None

        BaseSweep.kill(self)

    def step_param(self):
        """Iterates the parameter and checks for our stop condition.

        Returns:
        ---------
        Data pair in form of (<set param>, <setpoint>), or None if we have reached the end of our sweep.
        """
        # The AMI Magnet sweeps very slowly, so we implement sweeps of it  differently
        # If we are sweeping the magnet, let's deal with it here
        if isinstance(self.instrument, AMI430) and "field" in self.set_param.full_name:
            return self.step_AMI430()
        elif isinstance(self.instrument, M4G) and "field" in self.set_param.full_name:
            return self.step_M4G()

        # Check if we should continue (not at end yet)
        at_end = not (
            abs(self.setpoint - self.end) - abs(self.step / 2)
            > abs(self.step) * self.err
        )

        if not at_end:
            next_setpoint = self.setpoint + self.step
            next_setpoint = self._snap_to_step(next_setpoint, self._snap_origin, self.step)

            # If next setpoint exceeds bounds, treat as end of sweep
            sweep_min = min(self.begin, self.end)
            sweep_max = max(self.begin, self.end)
            if next_setpoint < sweep_min or next_setpoint > sweep_max:
                at_end = True
            else:
                self.setpoint = next_setpoint
                if not self.try_set(self.set_param, self.setpoint):
                    return None
                return [(self.set_param, self.setpoint)]

        # End of sweep handling
        if self.continuous:
            self.flip_direction()
            return self.step_param()
        elif self.bidirectional and self.direction == 0:
            self.flip_direction()
            return self.step_param()
        # If neither of the above are triggered, it means we are at the end of the sweep
        else:
            self.print_main.emit(
                f"Finished the sweep! {self.set_param.label} = {safe_get(self.set_param)} "
                f"({self.set_param.unit})"
            )
            self.flip_direction()
            return None

    def step_AMI430(self):
        """Used to control sweeps of AMI430 Instrument.

        The endpoint is determined prior to the sweep, and the current
        field is measured during ramping.

        Returns:
        ---------
        The parameter-value pair that was measured.
        """
        if self._ami_completion_pending:
            self._ami_completion_pending = False
            return None

        # Check if we have set the magnetic field yet
        if self.magnet_initialized is False:
            self.instrument.set_field(self.end, block=False)
            self.magnet_initialized = True
            time.sleep(self.inter_delay)

            while safe_get(self.instrument.ramping_state) != "ramping":
                time.sleep(self.inter_delay)

        # Grab our data
        dt = safe_get(self.set_param)
        data_pair = (self.set_param, dt)
        self.setpoint = dt

        # Check our stop conditions- being at the end point
        if safe_get(self.instrument.ramping_state) == "holding":
            self.magnet_initialized = False
            self.print_main.emit(
                f"Done with the sweep, {self.set_param.label}={self.set_param.get()} "
                f"({self.set_param.unit})"
            )
            self.flip_direction()
            self._ami_completion_pending = True

        # Return our data pair, just like any other sweep
        return [data_pair]

    def step_M4G(self, _retry_count=0):
        """Function to deal with sweeps of Cryomagnetics M4G Magnet Supply. Instead of setting intermediate points, we set the endpoint at the
        beginning, then ask it for the current field while it is ramping.
        """
        MAX_M4G_RETRIES = 100

        if self._m4g_completion_pending:
            self._m4g_completion_pending = False
            return None

        # Check if we have set the magnetic field yet
        if self.magnet_initialized is False:
            self.print_main.emit("Checking the remote connection to the magnet supply.")
            self.instrument.is_remote()
            time.sleep(1)

            # Start the field sweep
            self.instrument.field(self.end)
            self.magnet_initialized = True
        else:
            self.instrument.field(self.end)

        # Grab our data with proper error handling
        try:
            dt = safe_get(self.set_param)
            data_pair = (self.set_param, dt)
            self.setpoint = dt
        except Exception as e:
            if _retry_count >= MAX_M4G_RETRIES:
                error_msg = (
                    f"M4G field read failed after {MAX_M4G_RETRIES} retries: {e}"
                )
                self.print_main.emit(error_msg)
                self.mark_error(error_msg)
                return None
            self.print_main.emit(f"M4G read error (retry {_retry_count + 1}/{MAX_M4G_RETRIES}): {e}")
            time.sleep(self.inter_delay)
            return self.step_M4G(_retry_count=_retry_count + 1)
        # Check our stop conditions- being at the end point
        if (
            abs(self.instrument.field() - self.end) <= 0.001
        ):  # Check to see if field is within 1 mT of final value
            time.sleep(0.2)
            if (
                abs(self.instrument.field() - self.end) <= 0.001
            ):  # Wait for a little bit, then check again, if true then we are done with the sweep
                self.print_main.emit(
                    f"Done with the sweep, {self.set_param.label} = {self.set_param.get()} (T)"
                )

                if self.continuous:
                    self.flip_direction()
                    return self.step_M4G()
                elif self.bidirectional and self.direction == 0:
                    self.flip_direction()
                    return self.step_M4G()
                else:
                    self.magnet_initialized = False
                    self._m4g_completion_pending = True

        return [data_pair]

    def flip_direction(self):
        """Flips the direction of the sweep, to do bidirectional sweeping."""
        # If backwards, go forwards, and vice versa
        if self.direction:
            self.direction = 0
            self.step /= self.back_multiplier
        else:
            self.direction = 1
            self.step *= self.back_multiplier

        self.send_updates(no_sp=True)

        temp = self.begin
        self.begin = self.end
        self.end = temp
        self.step = -1 * self.step

        # Update snap origin to new begin on every flip. This ensures the step
        # grid always includes the current sweep's starting point.
        self._snap_origin = self.begin

        self.setpoint -= self.step
        self.setpoint = self._snap_to_step(self.setpoint, self._snap_origin, self.step)

        if self.plot_data is True and self.plotter is not None:
            self.add_break.emit(self.direction)

    def ramp_to(self, value, start_on_finish=False, persist=None, multiplier=1):
        """Ramps the set_param to a given value, at a rate dictated by the multiplier.

        Parameters
        ---------
        value:
            The desired starting value for the sweep.
        start_on_finish:
            Flag to determine whether to begin the sweep as soon as ramp is finished.
        multiplier:
            Factor to alter the step size, used to ramp quicker than the sweep speed.
        """
        # Ensure we aren't currently running
        if self.progressState.state == SweepState.RAMPING:
            self.print_main.emit(
                "Currently ramping. Finish current ramp before starting another."
            )
            return
        if self.progressState.state == SweepState.RUNNING:
            self.print_main.emit("Already running. Stop the sweep before ramping.")
            return

        # Check if we are already at the value
        curr_value = safe_get(self.set_param)
        if abs(value - curr_value) - abs(self.step / 2) < abs(self.step) * self.err:
            # self.print_main.emit(f"Already within {self.step} of the desired ramp value. Current value: {curr_value},
            # ramp setpoint: {value}.\nSetting our setpoint directly to the ramp value.")
            self.done_ramping(value, start_on_finish, persist)
            return

        # Create a new sweep to ramp our outer parameter to zero
        self.ramp_sweep = Sweep1D(
            self.set_param,
            curr_value,
            value,
            multiplier * self.step,
            inter_delay=self.inter_delay,
            complete_func=partial(self.done_ramping, value, start_on_finish, persist),
            save_data=False,
            plot_data=self.plot_data,
        )
        self.ramp_sweep.follow_param(self._params)

        self.progressState.state = SweepState.RAMPING
        self.ramp_sweep.start(ramp_to_start=False)

        self.print_main.emit(
            f"Ramping {self.set_param.label} to {value} ({self.set_param.unit}) . . . "
        )

    def ramp_to_zero(self):
        """Ramps the set_param to 0, at the same rate as already specified."""
        self.end = 0
        if self.setpoint - self.end > 0:
            self.step = (-1) * abs(self.step)
        else:
            self.step = abs(self.step)

        self.print_main.emit(f"Ramping {self.set_param.label} to 0 . . . ")
        self.start()

    @pyqtSlot()
    def done_ramping(self, value, start_on_finish=False, pd=None):
        """Alerts the sweep that the ramp is finished.

        Parameters
        ---------
        value:
            The starting value for the sweep, also the current parameter value.
        start_on_finish:
            Sweep will be called to start immediately after ramping when set to True.
        pd:
            Sets persistent data if running Sweep2D.
        """
        # Check if the ramp sweep itself failed with an error
        if self.ramp_sweep is not None and self.ramp_sweep.progressState.state == SweepState.ERROR:
            error_msg = self.ramp_sweep.progressState.error_message or "Ramp sweep failed"
            self.print_main.emit(f"Ramp sweep failed: {error_msg}")
            self.ramp_sweep.kill()
            self.ramp_sweep = None
            self.mark_error(f"Ramp to start failed: {error_msg}")
            return

        if self.progressState.state != SweepState.KILLED:
            self.progressState.state = SweepState.READY
        # Grab the beginning
        # value = self.ramp_sweep.begin

        # Check if we are at the value we expect, otherwise something went wrong with the ramp
        actual_value = safe_get(self.set_param)
        position_error = abs(actual_value - value) - abs(self.step / 2)
        tolerance = abs(self.step) * self.err
        if position_error > tolerance:
            error_msg = (
                f"Ramping failed (possible that the direction was changed while ramping). "
                f"Expected {self.set_param.label} final value: {value}. Actual value: "
                f"{actual_value}. Error: {position_error:.6g}, Tolerance: {tolerance:.6g}. "
                f"If tolerance is too tight, consider increasing the 'err' parameter (current: {self.err})."
            )
            self.print_main.emit(error_msg)

            if self.ramp_sweep is not None:
                self.ramp_sweep.kill()
                self.ramp_sweep = None

            self.mark_error(error_msg)
            return

        self.print_main.emit(
            f"Done ramping {self.set_param.label} to {value} ({self.set_param.unit})"
        )
        if not self.try_set(self.set_param, value):
            if self.ramp_sweep is not None:
                self.ramp_sweep.kill()
                self.ramp_sweep = None
            return
        self.setpoint = value - self.step
        self.setpoint = self._snap_to_step(self.setpoint, self._snap_origin, self.step)
        # if self.ramp_sweep is not None and self.ramp_sweep.plotter is not None:
        #    self.ramp_sweep.plotter.clear()

        if self.ramp_sweep is not None:
            self.ramp_sweep.kill()
            self.ramp_sweep = None

        if start_on_finish is True:
            self.start(persist_data=pd, ramp_to_start=False)

    def get_param_setpoint(self):
        """Obtains the current value of the setpoint."""
        return f"{self.set_param.label} = {safe_get(self.set_param)} ({self.set_param.unit})"

    def reset(self, new_params=None):
        """Resets Sweep1D, optional argument to change sweep details.

        Parameters
        ---------
        new_params:
            A list of 4 values to determine how we sweep.
            [ start value, stop value, step, frequency ]
        """
        # Set our new values if desired
        if new_params is not None:
            self.begin = new_params[0]
            self.end = new_params[1]
            self.step = new_params[2]
            self.inter_delay = 1 / new_params[3]
            # Update snap origin for the new grid
            self._snap_origin = new_params[0]

        # Reset our setpoint
        self.setpoint = self.begin - self.step
        self.setpoint = self._snap_to_step(self.setpoint, self._snap_origin, self.step)

        # Reset our plots
        self.plotter = None
        self.runner = None

    def estimate_time(self, verbose=True):
        """Estimate remaining time from the current sweep state."""
        if self.progressState.state == SweepState.DONE:
            remaining = 0
        else:
            if isinstance(self.instrument, AMI430):
                rate = safe_get(self.instrument.ramp_rate)
                if not rate:
                    return 0.0
                units = safe_get(self.instrument.ramp_rate_units)
                current = self.setpoint
                target = self.end
                distance = abs(target - current)
                if units == 0:
                    remaining = distance / rate
                else:
                    remaining = distance * 60 / rate

                if self.bidirectional and self.direction == 0:
                    distance_back = abs(self.end - self.begin)
                    if units == 0:
                        remaining += distance_back / rate
                    else:
                        remaining += distance_back * 60 / rate
                return remaining

            if isinstance(self.instrument, M4G):
                try:
                    rate = abs(float(self.instrument.range0_rate()))
                except Exception:
                    return 0.0
                if rate <= 0:
                    return 0.0
                current_value = self.setpoint
                distance = abs(self.end - current_value)
                remaining = distance / rate
                if self.bidirectional and self.direction == 0:
                    remaining += abs(self.end - self.begin) / rate
                return remaining

            current_value = self.setpoint

            step_size = abs(self.step)
            if step_size == 0:
                return 0.0

            distance = abs(self.end - current_value)
            remaining = (distance / step_size) * self.inter_delay

            if self.bidirectional and self.direction == 0:
                effective_back_multiplier = (
                    self.back_multiplier if self.back_multiplier not in (None, 0) else 1.0
                )
                if effective_back_multiplier > 0:
                    distance_back = abs(self.end - self.begin)
                    remaining += (
                        distance_back / (step_size * effective_back_multiplier)
                    ) * self.inter_delay

        if verbose:
            hours, minutes, seconds = self._split_hms(remaining)
            self.print_main.emit(
                f"Estimated time remaining for {repr(self)}: {hours}h:{minutes:02d}m:{seconds:02d}s"
            )

        return remaining

    # --- JSON export/import hooks ---
    def _export_json_specific(self, json_dict: dict) -> dict:
        json_dict["set_param"] = {
            "param": self.set_param.name,
            "instr_module": self.set_param.instrument.__class__.__module__,
            "instr_class": self.set_param.instrument.__class__.__name__,
            "instr_name": self.set_param.instrument.name,
            "start": self.begin,
            "stop": self.end,
            "step": self.step,
        }
        json_dict["attributes"]["bidirectional"] = self.bidirectional
        json_dict["attributes"]["continual"] = self.continuous
        json_dict["attributes"]["x_axis_time"] = self.x_axis
        return json_dict

    @classmethod
    def from_json(cls, json_dict, station):
        attrs = dict(json_dict.get("attributes", {}))
        bidirectional = attrs.pop("bidirectional", False)
        continual = attrs.pop("continual", False)
        x_axis_time = attrs.pop("x_axis_time", 0)

        sp = json_dict["set_param"]
        p = BaseSweep._load_parameter_by_type(
            sp["param"],
            sp["instr_name"],
            sp["instr_module"],
            sp["instr_class"],
            station,
        )
        return cls(
            p,
            sp["start"],
            sp["stop"],
            sp["step"],
            bidirectional=bidirectional,
            continual=continual,
            x_axis_time=x_axis_time,
            **attrs,
        )

    def _params_to_exclude_from_follow(self) -> set:
        """Exclude the set_param from follow_params export for 1D sweeps."""
        try:
            return {f"{self.set_param.instrument.name}.{self.set_param.name}"}
        except Exception:
            return set()

    def __del__(self):
        """Destructor. Should delete all child threads and close all figures when the sweep object is deleted."""
        self.kill()



# heatmap_thread.py

import math
from collections import deque

import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.axes_grid1 import make_axes_locatable
from PyQt5.QtCore import QObject, pyqtSlot


class Heatmap(QObject):  # moved to legacy
    """Thread to control the plotting for Sweep2D.

    Gathers data from the RunnerThread. For Sweep2D, in addition to plotting
    the followed parameters against the set parameter, the heatmap thread
    creates a 3D representation of the followed parameter against an inner
    and outer set parameter, using color to represent the third dimension.

    Attributes:
    ---------
    sweep:
        The parent sweep object.
    lines_to_add:
        Stores Line2D objects to be added to the heatmap.
    count:
        Used to index the heatmap plotting dictionary.
    max_datapt:
        The maximum value of the 'set_param' data (y).
    min_datapt:
        The minimum value of the 'set_param' data (y).
    figs_set:
        Changes to true after figures have been created.

    Methods:
    ---------
    create_figs()
        Creates the heatmap figure for Sweep2D.
    add_lines(lines)
        Adds Line2D objects to be plotted on the heatmap.
    add_to_plot(line)
        Plots the Line2D objects set to be added.
    update_data(x_out)
        Updates outer parameter data for all inner parameters.
    update_heatmap()
        Prepares lines to be added for plotting.
    clear()
        Closes all active figures.
    """

    def __init__(self, sweep):
        """Initializes the thread.

        Takes in the parent sweep and the figure information
        to use previously obtained plot if desired.
        """
        self.sweep = sweep
        # Datastructure to
        self.lines_to_add = deque([])
        self.count = 0
        self.max_datapt = float("-inf")
        self.min_datapt = float("inf")
        self.figs_set = False

        QObject.__init__(self)

    def create_figs(self):
        """Creates the heatmap figure for the 2D sweep.

        If parent sweep has been previously plotted, this thread will use
        the previously created figures.
        """
        if self.figs_set is True:
            return

        # First, determine the resolution on each axis
        self.res_in = (
            math.ceil(
                abs((self.sweep.in_stop - self.sweep.in_start) / self.sweep.in_step)
            )
            + 1
        )
        self.res_out = (
            math.ceil(
                abs((self.sweep.out_stop - self.sweep.out_start) / self.sweep.out_step)
            )
            + 1
        )

        # Create the heatmap data matrix - initially as all 0s
        self.heatmap_dict = {}
        self.out_keys = set([])
        self.in_keys = set([])
        self.out_step = self.sweep.out_step
        self.in_step = self.sweep.in_step
        for x_out in np.linspace(
            self.sweep.out_start,
            self.sweep.out_stop,
            int(
                abs(self.sweep.out_stop - self.sweep.out_start)
                / abs(self.sweep.out_step)
                + 1
            ),
            endpoint=True,
        ):
            self.heatmap_dict[x_out] = {}
            self.out_keys.add(x_out)
            for x_in in np.linspace(
                self.sweep.in_start,
                self.sweep.in_stop,
                int(
                    abs(self.sweep.in_stop - self.sweep.in_start)
                    / abs(self.sweep.in_step)
                    + 1
                ),
                endpoint=True,
            ):
                self.heatmap_dict[x_out][x_in] = 0
                self.in_keys.add(x_in)
        self.out_keys = sorted(self.out_keys)
        self.in_keys = sorted(self.in_keys)

        self.heatmap_data = np.zeros((self.res_out, self.res_in))
        # Create a figure
        self.heat_fig = plt.figure()
        # Use plt.imshow to actually plot the matrix
        self.heatmap = plt.imshow(self.heatmap_data)
        ax = plt.gca()
        self.heat_ax = ax

        # Set our axes and ticks
        plt.ylabel(f"{self.sweep.set_param.label} ({self.sweep.set_param.unit})")
        plt.xlabel(f"{self.sweep.in_param.label} ({self.sweep.in_param.unit})")
        inner_tick_lbls = np.linspace(self.sweep.in_start, self.sweep.in_stop, 5)
        outer_tick_lbls = np.linspace(self.sweep.out_stop, self.sweep.out_start, 5)

        ax.set_xticks(np.linspace(0, self.res_in - 1, 5))
        ax.set_yticks(np.linspace(0, self.res_out - 1, 5))
        ax.set_xticklabels(inner_tick_lbls)
        ax.set_yticklabels(outer_tick_lbls)

        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)

        # Create our colorbar scale
        cbar = plt.colorbar(self.heatmap, cax=cax)
        # Create label
        plot_para = self.sweep.in_sweep._params[self.sweep.heatmap_ind]
        cbar.set_label(f"{plot_para.label} ({plot_para.unit})")
        self.figs_set = True

    @pyqtSlot(list)
    def add_lines(self, lines):
        """Feeds the thread Line2D objects to add to the heatmap.

        Parameters
        ---------
        lines:
            A  dictionary of Line2D objects (backwards and forwards)
            to be added to the heatmap.
        """
        self.lines_to_add.append(lines)
        try:
            self.update_heatmap()
        except Exception as e:
            print("Failed to update the heatmap: ", e)

    def add_to_plot(self, line):
        """Plots forward Line2D objects from lines dictionary ('add_lines').

        Parameters
        ---------
        line:
            Set in 'update_heatmap' as the forward line to be added; each point
            of these lines represents two independent parameter values.
        """
        in_key = 0

        x_raw, y_raw = line.get_data()

        x_data = [i for i in x_raw if not math.isnan(i)]
        y_data = [i for i in y_raw if not math.isnan(i)]

        for key in self.in_keys:
            if abs(x_data[0] - key) < abs(self.in_step / 2):
                in_key = key

        start_pt = self.in_keys.index(in_key)

        for i, x in enumerate(x_data):
            self.heatmap_dict[self.out_keys[self.count]][self.in_keys[start_pt + i]] = (
                y_data[i]
            )
            if y_data[i] > self.max_datapt:
                self.max_datapt = y_data[i]
            if y_data[i] < self.min_datapt:
                self.min_datapt = y_data[i]
        self.update_data(self.res_out - self.count - 1)
        self.count += 1
        # print(f"called heatmap from thread: {QThread.currentThreadId()}")

    def update_data(self, x_out):
        """Updates outer parameter data for all inner parameters."""
        for i, x in enumerate(self.in_keys):
            self.heatmap_data[x_out][i] = self.heatmap_dict[self.out_keys[self.count]][
                x
            ]

    def update_heatmap(self):
        """Prepares lines to be added for plotting."""
        while len(self.lines_to_add) != 0:
            # Grab the lines to add
            line_pair = self.lines_to_add.popleft()

            forward_line = line_pair[0]
            backward_line = line_pair[1]

            self.add_to_plot(forward_line)

        # Refresh the image!
        self.heatmap.set_data(self.heatmap_data)
        self.heatmap.set_clim(self.min_datapt, self.max_datapt)
        self.heat_fig.canvas.draw()
        self.heat_fig.canvas.flush_events()

        # Smart sleep, by checking if the whole process has taken longer than
        # our sleep time
        #    sleep_time = self.sweep.inter_delay - (time.monotonic() - t)
        #    if sleep_time > 0:
        #        time.sleep(sleep_time)

    def clear(self):
        """Closes all active figures."""
        plt.close(self.heat_fig)
        self.figs_set = False



import time

import numpy as np
from PyQt5.QtCore import QObject

from ..tools.util import _autorange_srs, safe_get
from .progress import SweepState
from .sweep1d import Sweep1D


class GateLeakage(Sweep1D, QObject):
    """Extension of Sweep1D to perform gate leakage measurements.

    Sweeps 'set_param' (voltage) while monitoring 'track_param' (current) with
    noise-immune leak detection: triggers on 2 consecutive points exceeding max_I,
    clears after 5 consecutive safe points.

    Attributes:
    ---------
    set_param:
        The independent variable to be swept through a desired range.
    track_param:
        The followed parameter to be measured during the sweep.
    max_I:
        Maximum allowed absolute value for track_param. Sweep reverses when
        |track_param| >= max_I for 2 consecutive points.
    step:
        The step size for the independent variable sweep.
    limit:
        The end value for Sweep1D. Defaults to infinity.
    start:
        The start value for Sweep1D. Defaults to zero.
    flips:
        Tracks the number of times the sweep has changed direction.
    leak_trigger_count:
        Consecutive points where |track_param| >= max_I.
    safe_trigger_count:
        Consecutive points where |track_param| < max_I.
    leak_detected:
        True when leak confirmed (2 consecutive triggers), clears after 5 safe points.

    Methods:
    ---------
    step_param()
        Runs sweep once in both directions while measuring 'track_param'.
    update_values()
        Keeps record of independent variable data during sweep.
    flip_direction()
        Changes direction of the sweep.
    """

    def __init__(
        self, set_param, track_param, max_I, step, limit=np.inf, start=0, **kwargs
    ):
        self.max_I = max_I
        self.flips = 0
        self.track_param = track_param
        self.leak_trigger_count = 0
        self.safe_trigger_count = 0
        self.leak_detected = False

        Sweep1D.__init__(self, set_param, start, limit, step, **kwargs)
        QObject.__init__(self)

        self.follow_param(self.track_param)

    def step_param(self):
        """Runs Sweep1D in both directions by step size.

        Stores data of followed 'track_param' at each setpoint of 'set_param'.
        Changes direction when: (1) |track_param| >= max_I for 2 consecutive points,
        or (2) set_param reaches its limit. Leak detection clears after 5 consecutive
        safe points (|track_param| < max_I). Sweep ends after two direction changes.

        Returns:
        ---------
        A list containing the values of 'set_param' and 'track_param' for each
        setpoint until the sweep is stopped after 2 total flips.
        """
        # Our ending condition is if we end up back at 0 after going forwards and backwards
        if self.flips >= 2 and abs(self.setpoint) <= abs(self.step / (3 / 2)):
            self.flips = 0
            print(f"Done with the sweep, {self.set_param.label}={self.setpoint}")
            # Return None to signal completion without emitting a synthetic datapoint
            # that could violate parameter validators or pollute the dataset
            return None

        # Check if we're at the end (within half a step)
        at_end = abs(self.end) != np.inf and abs(self.setpoint - self.end) <= abs(self.step)

        if at_end:
            self.flip_direction()
            print("tripped output limit")
            return self.step_param()

        # Compute next setpoint and check bounds before setting
        next_setpoint = self.setpoint + self.step
        next_setpoint = self._snap_to_step(next_setpoint, self._snap_origin, self.step)

        # Check if next setpoint exceeds sweep bounds
        if abs(self.end) != np.inf:
            sweep_min = min(self.begin, self.end)
            sweep_max = max(self.begin, self.end)
            if next_setpoint < sweep_min or next_setpoint > sweep_max:
                # Would exceed bounds, treat as end of sweep
                self.flip_direction()
                print("tripped output limit (bounds check)")
                return self.step_param()

        self.setpoint = next_setpoint
        if not self.try_set(self.set_param, self.setpoint):
            return None

        v = safe_get(self.track_param)

        # Check if current exceeds threshold
        if abs(v) >= abs(self.max_I):
            self.leak_trigger_count += 1
            self.safe_trigger_count = 0

            if self.leak_trigger_count >= 2 and not self.leak_detected:
                self.leak_detected = True
                self.flip_direction()
                # After flip, compute and check next setpoint before setting
                post_flip_setpoint = self.setpoint + self.step
                post_flip_setpoint = self._snap_to_step(post_flip_setpoint, self._snap_origin, self.step)
                # Only set if within bounds
                if abs(self.end) != np.inf:
                    sweep_min = min(self.begin, self.end)
                    sweep_max = max(self.begin, self.end)
                    if sweep_min <= post_flip_setpoint <= sweep_max:
                        self.setpoint = post_flip_setpoint
                else:
                    self.setpoint = post_flip_setpoint
                print("tripped input limit")
        else:
            self.safe_trigger_count += 1
            self.leak_trigger_count = 0

            if self.safe_trigger_count >= 5 and self.leak_detected:
                self.leak_detected = False

        return [(self.set_param, self.setpoint), (self.track_param, v)]

    def update_values(self):
        """Iterates and keeps record of independent variable data during sweep.

        Iterates each 'set_param' individually beginning with time.
        If we are saving data, it happens here, and the data is returned.

        Returns:
        ---------
        data:
            A list of tuples with the new data. Each tuple is of the format
            (<QCoDeS Parameter>, measurement value). The tuples are passed in order of
            time, then set_param (if applicable), then all the followed params.
        """
        t = time.monotonic() - self.t0

        data = []
        data.append(("time", t))

        step_data = self.step_param()
        if step_data is None:
            if self.progressState.state != SweepState.KILLED:
                self.mark_done()
            return None

        data += step_data

        persist_param = None
        if self.persist_data is not None:
            data.append(self.persist_data)
            persist_param = self.persist_data[0]

        for i, (l, _, gain) in enumerate(self._srs):
            _autorange_srs(l, 3)

        for i, p in enumerate(self._params):
            if p is not persist_param and p is not self.track_param:
                v = safe_get(p)
                data.append((p, v))

        if self.save_data and self.progressState.state == SweepState.RUNNING:
            self.runner.datasaver.add_result(*data)

        self.send_updates()

        return data

    def flip_direction(self):
        """Changes direction of the sweep.

        Note: Unlike Sweep1D, GateLeakage sweeps from a center point (start, default 0)
        outward in both directions. After each flip, _snap_origin is updated to the new
        begin to ensure the step grid includes the current sweep's starting point.
        """
        self.flips += 1

        if self.flips >= 2 and self.setpoint > 0:
            self.step = (-1) * abs(self.step)
            self.end = (-1) * self.end
            self.begin = self.setpoint  # New sweep starts from current position
            self._snap_origin = self.begin
        elif self.flips >= 2 and self.setpoint < 0:
            self.step = abs(self.step)
            self.end = (-1) * self.end
            self.begin = self.setpoint  # New sweep starts from current position
            self._snap_origin = self.begin
        elif self.flips < 2:
            self.end = (-1) * self.end
            self.step = -1 * self.step
            self.begin = self.setpoint  # New sweep starts from current position
            self._snap_origin = self.begin
            self.setpoint -= self.step
            self.setpoint = self._snap_to_step(self.setpoint, self._snap_origin, self.step)

        # If backwards, go forwards, and vice versa
        if self.direction:
            self.direction = 0
        else:
            self.direction = 1

    def estimate_time(self, verbose=True):
        """Gate leakage sweeps are event-driven; report no deterministic estimate."""
        if verbose:
            self.print_main.emit(
                f"No estimated time remaining for {repr(self)} (non-deterministic sweep)."
            )
        return 0.0
